# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=missing-docstring,no-self-use,no-init,invalid-name
"""Unit tests for the path_template module."""

from __future__ import absolute_import
import unittest2

from google.api.control.path_template import PathTemplate, ValidationException


class TestPathTemplate(unittest2.TestCase):
    """Unit tests for PathTemplate."""

    def test_len(self):
        self.assertEqual(len(PathTemplate('a/b/**/*/{a=hello/world}')), 6)

    def test_fail_invalid_token(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'hello/wor*ld')

    def test_fail_when_impossible_match(self):
        template = PathTemplate('hello/world')
        self.assertRaises(ValidationException,
                          template.match, 'hello')
        template = PathTemplate('hello/world')
        self.assertRaises(ValidationException,
                          template.match, 'hello/world/fail')

    def test_fail_mismatched_literal(self):
        template = PathTemplate('hello/world')
        self.assertRaises(ValidationException,
                          template.match, 'hello/world2')

    def test_fail_when_multiple_path_wildcards(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'buckets/*/**/**/objects/*')

    def test_fail_if_inner_binding(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'buckets/{hello={world}}')

    def test_fail_unexpected_eof(self):
        self.assertRaises(ValidationException,
                          PathTemplate, 'a/{hello=world')

    def test_match_atomic_resource_name(self):
        template = PathTemplate('buckets/*/*/objects/*')
        self.assertEqual({'$0': 'f', '$1': 'o', '$2': 'bar'},
                         template.match('buckets/f/o/objects/bar'))
        template = PathTemplate('/buckets/{hello}')
        self.assertEqual({'hello': 'world'},
                         template.match('buckets/world'))
        template = PathTemplate('/buckets/{hello=*}')
        self.assertEqual({'hello': 'world'},
                         template.match('buckets/world'))

    def test_match_escaped_chars(self):
        template = PathTemplate('buckets/*/objects')
        self.assertEqual({'$0': 'hello%2F%2Bworld'},
                         template.match('buckets/hello%2F%2Bworld/objects'))

    def test_match_template_with_unbounded_wildcard(self):
        template = PathTemplate('buckets/*/objects/**')
        self.assertEqual({'$0': 'foo', '$1': 'bar/baz'},
                         template.match('buckets/foo/objects/bar/baz'))

    def test_match_with_unbound_in_middle(self):
        template = PathTemplate('bar/**/foo/*')
        self.assertEqual({'$0': 'foo/foo', '$1': 'bar'},
                         template.match('bar/foo/foo/foo/bar'))

    def test_render_atomic_resource(self):
        template = PathTemplate('buckets/*/*/*/objects/*')
        url = template.render({
            '$0': 'f', '$1': 'o', '$2': 'o', '$3': 'google.com:a-b'})
        self.assertEqual(url, 'buckets/f/o/o/objects/google.com')

    def test_render_fail_when_too_few_variables(self):
        template = PathTemplate('buckets/*/*/*/objects/*')
        self.assertRaises(ValidationException,
                          template.render,
                          {'$0': 'f', '$1': 'l', '$2': 'o'})

    def test_render_with_unbound_in_middle(self):
        template = PathTemplate('bar/**/foo/*')
        url = template.render({'$0': '1/2', '$1': '3'})
        self.assertEqual(url, 'bar/1/2/foo/3')

    def test_to_string(self):
        template = PathTemplate('bar/**/foo/*')
        self.assertEqual(str(template), 'bar/{$0=**}/foo/{$1=*}')
        template = PathTemplate('buckets/*/objects/*')
        self.assertEqual(str(template), 'buckets/{$0=*}/objects/{$1=*}')
        template = PathTemplate('/buckets/{hello}')
        self.assertEqual(str(template), 'buckets/{hello=*}')
        template = PathTemplate('/buckets/{hello=what}/{world}')
        self.assertEqual(str(template), 'buckets/{hello=what}/{world=*}')
        template = PathTemplate('/buckets/helloazAZ09-.~_what')
        self.assertEqual(str(template), 'buckets/helloazAZ09-.~_what')



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import datetime
import unittest2
from expects import be_none, be_true, expect, equal, raise_error

import google.api.gen.servicecontrol_v1_messages as messages
from google.api.control import (distribution, metric_descriptor, timestamp, MetricKind,
                        ValueType)
from google.api.control import metric_value, operation, report_request


_KNOWN = metric_descriptor.KnownMetrics


def _wanted_distribution_with_sample(value, *args):
    d = distribution.create_exponential(*args)
    distribution.add_sample(value, d)
    return d


def _given_info(wanted_size, test_api_key, api_key_valid=True):
    return report_request.Info(
        request_size=wanted_size,
        response_size=wanted_size,
        request_time=datetime.timedelta(seconds=3),
        backend_time=datetime.timedelta(seconds=2),
        overhead_time=datetime.timedelta(seconds=1),
        api_key=test_api_key,
        api_key_valid=api_key_valid,
    )


class KnownMetricsBase(object):
    SUBJECT = None
    WANTED_ADDED_METRICS = []
    WANTED_SIZE = 7426 # arbitrary
    TEST_API_KEY = 'test_key'
    GIVEN_INFO = _given_info(WANTED_SIZE, TEST_API_KEY, api_key_valid=True)

    def _base_operation(self):
        return messages.Operation(
            consumerId='project:project_id',
            operationId='an_op_id',
            operationName='an_op_name')

    def _wanted_operation(self):
        op = self._base_operation()
        if self.WANTED_ADDED_METRICS:
            op.metricValueSets.append(self.WANTED_ADDED_METRICS)
        return op

    def _matching_descriptor(self):
        return messages.MetricDescriptor(
            name=self.SUBJECT.metric_name,
            metricKind=self.SUBJECT.kind,
            valueType=self.SUBJECT.value_type)

    def _not_matched(self):
        d = self._matching_descriptor()
        d.metricKind = MetricKind.METRIC_KIND_UNSPECIFIED
        return d

    def test_should_be_supported(self):
        expect(_KNOWN.is_supported(self._matching_descriptor())).to(be_true)
        expect(_KNOWN.is_supported(self._not_matched())).not_to(be_true)

    def test_should_be_matched_correctly(self):
        expect(self.SUBJECT.matches(self._matching_descriptor())).to(be_true)
        expect(self.SUBJECT.matches(self._not_matched())).not_to(be_true)

    def test_should_update_request_info(self):
        an_op = self._base_operation()
        wanted_op = self._wanted_operation()
        self.SUBJECT.do_operation_update(self.GIVEN_INFO, an_op)
        expect(an_op).to(equal(wanted_op))


class KnownMetricsInvalidApiKey(KnownMetricsBase):
    GIVEN_INFO = _given_info(KnownMetricsBase.WANTED_SIZE,
                             KnownMetricsBase.TEST_API_KEY,
                             api_key_valid=False)


class TestConsumerRequestCount(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_REQUEST_COUNT
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(int64Value=1)])


class TestConsumerRequestCountInvalidApiKey(KnownMetricsInvalidApiKey,
                                            unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_REQUEST_COUNT


class TestProducerRequestCount(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_REQUEST_COUNT
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(int64Value=1)])


class TestConsumerRequestSizes(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_REQUEST_SIZES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.WANTED_SIZE,
        *metric_descriptor._SIZE_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerRequestSizesInvalidApiKey(KnownMetricsInvalidApiKey,
                                            unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_REQUEST_SIZES


class TestProducerRequestSizes(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_REQUEST_SIZES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.WANTED_SIZE,
        *metric_descriptor._SIZE_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerResponseSizes(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_RESPONSE_SIZES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.WANTED_SIZE,
        *metric_descriptor._SIZE_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerResponseSizesInvalidApiKey(KnownMetricsInvalidApiKey,
                                             unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_RESPONSE_SIZES


class TestProducerResponseSizes(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_RESPONSE_SIZES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.WANTED_SIZE,
        *metric_descriptor._SIZE_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerErrorCountNoError(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_ERROR_COUNT


class TestConsumerErrorCountWithErrors(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_ERROR_COUNT
    GIVEN_INFO = report_request.Info(
        response_code=401,
        api_key=KnownMetricsBase.TEST_API_KEY,
        api_key_valid=True,
    )
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(int64Value=1)])


class TestConsumerErrorCountInvalidApiKey(KnownMetricsInvalidApiKey,
                                          unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_ERROR_COUNT


class TestProducerErrorCountNoError(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_ERROR_COUNT


class TestProducerErrorCountWithErrors(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_ERROR_COUNT
    GIVEN_INFO = report_request.Info(
        response_code=401,
        api_key=KnownMetricsBase.TEST_API_KEY,
        api_key_valid=True,
    )
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(int64Value=1)])


class TestConsumerTotalLatencies(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_TOTAL_LATENCIES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.GIVEN_INFO.request_time.seconds,
        *metric_descriptor._TIME_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerTotalLatenciesInvalidApiKey(KnownMetricsInvalidApiKey,
                                              unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_TOTAL_LATENCIES


class TestProducerTotalLatencies(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_TOTAL_LATENCIES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.GIVEN_INFO.request_time.seconds,
        *metric_descriptor._TIME_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerRequestOverheadLatencies(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_REQUEST_OVERHEAD_LATENCIES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.GIVEN_INFO.overhead_time.seconds,
        *metric_descriptor._TIME_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerRequestOverheadLatenciesInvalidApiKey(
    KnownMetricsInvalidApiKey, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_REQUEST_OVERHEAD_LATENCIES


class TestProducerRequestOverheadLatencies(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_REQUEST_OVERHEAD_LATENCIES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.GIVEN_INFO.overhead_time.seconds,
        *metric_descriptor._TIME_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerBackendLatencies(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_BACKEND_LATENCIES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.GIVEN_INFO.backend_time.seconds,
        *metric_descriptor._TIME_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])


class TestConsumerBackendLatenciesInvalidApiKey(KnownMetricsInvalidApiKey,
                                                unittest2.TestCase):
    SUBJECT = _KNOWN.CONSUMER_BACKEND_LATENCIES


class TestProducerBackendLatencies(KnownMetricsBase, unittest2.TestCase):
    SUBJECT = _KNOWN.PRODUCER_BACKEND_LATENCIES
    WANTED_DISTRIBUTION = _wanted_distribution_with_sample(
        KnownMetricsBase.GIVEN_INFO.backend_time.seconds,
        *metric_descriptor._TIME_DISTRIBUTION_ARGS)
    WANTED_ADDED_METRICS = messages.MetricValueSet(
        metricName=SUBJECT.metric_name,
        metricValues=[metric_value.create(distributionValue=WANTED_DISTRIBUTION)])



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""money provides funcs for working with `Money` instances.

:func:`check_valid` determines if a `Money` instance is valid
:func:`add` adds two `Money` instances together

"""

from __future__ import absolute_import

import logging
import sys

from . import messages

logger = logging.getLogger(__name__)

_INT64_MAX = sys.maxint
_INT64_MIN = -sys.maxint - 1
_BILLION = 1000000000
MAX_NANOS = _BILLION - 1
_MSG_3_LETTERS_LONG = 'The currency code is not 3 letters long'
_MSG_UNITS_NANOS_MISMATCH = 'The signs of the units and nanos do not match'
_MSG_NANOS_OOB = 'The nanos field must be between -999999999 and 999999999'


def check_valid(money):
    """Determine if an instance of `Money` is valid.

    Args:
      money (:class:`google.api.gen.servicecontrol_v1_messages.Money`): the
        instance to test

    Raises:
      ValueError: if the money instance is invalid
    """
    if not isinstance(money, messages.Money):
        raise ValueError('Inputs should be of type %s' % (messages.Money,))
    currency = money.currencyCode
    if not currency or len(currency) != 3:
        raise ValueError(_MSG_3_LETTERS_LONG)
    units = money.units
    nanos = money.nanos
    if ((units > 0) and (nanos < 0)) or ((units < 0) and (nanos > 0)):
        raise ValueError(_MSG_UNITS_NANOS_MISMATCH)
    if abs(nanos) > MAX_NANOS:
        raise ValueError(_MSG_NANOS_OOB)


def add(a, b, allow_overflow=False):
    """Adds two instances of `Money`.

    Args:
      a (:class:`google.api.gen.servicecontrol_v1_messages.Money`): one money
        value
      b (:class:`google.api.gen.servicecontrol_v1_messages.Money`): another
        money value
      allow_overflow: determines if the addition is allowed to overflow

    Return:
      `Money`: an instance of Money

    Raises:
      ValueError: if the inputs do not have the same currency code
      OverflowError: if the sum overflows and allow_overflow is not `True`
    """
    for m in (a, b):
        if not isinstance(m, messages.Money):
            raise ValueError('Inputs should be of type %s' % (messages.Money,))
    if a.currencyCode != b.currencyCode:
        raise ValueError('Money values need the same currency to be summed')
    nano_carry, nanos_sum = _sum_nanos(a, b)
    units_sum_no_carry = a.units + b.units
    units_sum = units_sum_no_carry + nano_carry

    # Adjust when units_sum and nanos_sum have different signs
    if units_sum > 0 and nanos_sum < 0:
        units_sum -= 1
        nanos_sum += _BILLION
    elif units_sum < 0 and nanos_sum > 0:
        units_sum += 1
        nanos_sum -= _BILLION

    # Return the result, detecting overflow if it occurs
    sign_a = _sign_of(a)
    sign_b = _sign_of(b)
    if sign_a > 0 and sign_b > 0 and units_sum >= _INT64_MAX:
        if not allow_overflow:
            raise OverflowError('Money addition positive overflow')
        else:
            return messages.Money(units=_INT64_MAX,
                                  nanos=MAX_NANOS,
                                  currencyCode=a.currencyCode)
    elif (sign_a < 0 and sign_b < 0 and
          (units_sum_no_carry <= -_INT64_MAX or units_sum <= -_INT64_MAX)):
        if not allow_overflow:
            raise OverflowError('Money addition negative overflow')
        else:
            return messages.Money(units=_INT64_MIN,
                                  nanos=-MAX_NANOS,
                                  currencyCode=a.currencyCode)
    else:
        return messages.Money(units=units_sum,
                              nanos=nanos_sum,
                              currencyCode=a.currencyCode)


def _sum_nanos(a, b):
    the_sum = a.nanos + b.nanos
    carry = 0
    if the_sum > _BILLION:
        carry = 1
        the_sum -= _BILLION
    elif the_sum <= -_BILLION:
        carry = -1
        the_sum += _BILLION
    return carry, the_sum


def _sign_of(money):
    """Determines the amount sign of a money instance

    Args:
      money (:class:`google.api.gen.servicecontrol_v1_messages.Money`): the
        instance to test

    Return:
      int: 1, 0 or -1

    """
    units = money.units
    nanos = money.nanos
    if units:
        if units > 0:
            return 1
        elif units < 0:
            return -1
    if nanos:
        if nanos > 0:
            return 1
        elif nanos < 0:
            return -1
    return 0



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""report_request supports aggregation of ReportRequests.

It proves :class:`.Aggregator` that aggregates and batches together
ReportRequests.

"""


from __future__ import absolute_import

import collections
import functools
import hashlib
import logging
import time
from datetime import datetime, timedelta

from apitools.base.py import encoding
from enum import Enum
from . import caches, label_descriptor, operation, messages
from . import metric_descriptor, signing, timestamp

logger = logging.getLogger(__name__)

SIZE_NOT_SET = -1


def _validate_int_arg(name, value):
    if value == SIZE_NOT_SET or (isinstance(value, int) and value > 0):
        return
    raise ValueError('%s should be a positive int/long' % (name,))


def _validate_timedelta_arg(name, value):
    if value is None or isinstance(value, timedelta):
        return
    raise ValueError('%s should be a timedelta' % (name,))


class ReportingRules(collections.namedtuple('ReportingRules',
                                            ['logs', 'metrics', 'labels'])):
    """Holds information that determines how to fill a `ReportRequest`.

    Attributes:
      logs (iterable[string]): the name of logs to be included in the `ReportRequest`
      metrics (iterable[:class:`google.api.control.metric_descriptor.KnownMetrics`]):
        the metrics to be added to a `ReportRequest`
      labels (iterable[:class:`google.api.control.metric_descriptor.KnownLabels`]):
        the labels to be added to a `ReportRequest`
    """
    # pylint: disable=too-few-public-methods

    def __new__(cls, logs=None, metrics=None, labels=None):
        """Invokes the base constructor with default values."""
        logs = set(logs) if logs else set()
        metrics = tuple(metrics) if metrics else tuple()
        labels = tuple(labels) if labels else tuple()
        return super(cls, ReportingRules).__new__(cls, logs, metrics, labels)

    @classmethod
    def from_known_inputs(cls, logs=None, metric_names=None, label_names=None):
        """An alternate constructor that assumes known metrics and labels.

        This differs from the default constructor in that the metrics and labels
        are iterables of names of 'known' metrics and labels respectively. The
        names are used to obtain the metrics and labels from
        :class:`google.api.control.metric_descriptor.KnownMetrics` and
        :class:`google.api.control.label_descriptor.KnownLabels` respectively.

        names that don't correspond to a known metric or label are ignored; as
        are metrics or labels that don't yet have a way of updating the
        `ReportRequest` operation.

        Args:
          logs (iterable[string]): the name of logs to be included in the
            `ReportRequest`
          metric_names (iterable[string]): the name of a known metric to be
            added to the `ReportRequest`
          label_names (iterable[string]): the name of a known label to be added
            to the `ReportRequest`

        """
        if not metric_names:
            metric_names = ()
        if not label_names:
            label_names = ()
        known_labels = []
        known_metrics = []
        # pylint: disable=no-member
        # pylint is not aware of the __members__ attributes
        for l in label_descriptor.KnownLabels.__members__.values():
            if l.update_label_func and l.label_name in label_names:
                known_labels.append(l)
        for m in metric_descriptor.KnownMetrics.__members__.values():
            if m.update_op_func and m.metric_name in metric_names:
                known_metrics.append(m)
        return cls(logs=logs, metrics=known_metrics, labels=known_labels)


class ReportedProtocols(Enum):
    """Enumerates the protocols that can be reported."""
    # pylint: disable=too-few-public-methods
    UNKNOWN = 0
    HTTP = 1
    HTTP2 = 2
    GRPC = 3


class ReportedPlatforms(Enum):
    """Enumerates the platforms that can be reported."""
    # pylint: disable=too-few-public-methods
    UNKNOWN = 0
    GAE_FLEX = 1
    GAE_STANDARD = 2
    GCE = 3
    GKE = 4
    DEVELOPMENT = 5


class ErrorCause(Enum):
    """Enumerates the causes of errors."""
    # pylint: disable=too-few-public-methods
    internal = 0  # default, error in scc library code
    application = 1  # external application error
    auth = 2  # authentication error
    service_control = 3  # error in service control check


# alias the severity enum
_SEVERITY = messages.LogEntry.SeverityValueValuesEnum


def _struct_payload_from(a_dict):
    return encoding.PyValueToMessage(messages.LogEntry.StructPayloadValue, a_dict)


class Info(
        collections.namedtuple(
            'Info', (
                'api_name',
                'api_method',
                'api_version',
                'auth_issuer',
                'auth_audience',
                'backend_time',
                'error_cause',
                'location',
                'log_message',
                'method',
                'overhead_time',
                'platform',
                'producer_project_id',
                'protocol',
                'request_size',
                'request_time',
                'response_code',
                'response_size',
                'url',
            ) + operation.Info._fields),
        operation.Info):
    """Holds the information necessary to fill in a ReportRequest.

    In the attribute descriptions below, N/A means 'not available'

    Attributes:
       api_name (string): the api name and version
       api_method (string): the full api method name
       api_version (string): the api version
       auth_issuer (string): the auth issuer
       auth_audience (string): the auth audience
       backend_time(datetime.timedelta): the backend request time, None for N/A
       error_cause(:class:`ErrorCause`): the cause of error if one has occurred
       location (string): the location of the service
       log_message (string): a message to log as an info log
       method (string): the HTTP method used to make the request
       overhead_time(datetime.timedelta): the overhead time, None for N/A
       platform (:class:`ReportedPlatform`): the platform in use
       producer_project_id (string): the producer project id
       protocol (:class:`ReportedProtocol`): the protocol used
       request_size(int): the request size in bytes, -1 means N/A
       request_time(datetime.timedelta): the request time
       response_size(int): the request size in bytes, -1 means N/A
       response_code(int): the code of the http response
       url (string): the request url

    """
    # pylint: disable=too-many-arguments,too-many-locals

    COPYABLE_LOG_FIELDS = [
        'api_name',
        'api_method',
        'api_key',
        'producer_project_id',
        'referer',
        'location',
        'log_message',
        'url',
    ]

    def __new__(cls,
                api_name='',
                api_method='',
                api_version='',
                auth_issuer='',
                auth_audience='',
                backend_time=None,
                error_cause=ErrorCause.internal,
                location='',
                log_message='',
                method='',
                overhead_time=None,
                platform=ReportedPlatforms.UNKNOWN,
                producer_project_id='',
                protocol=ReportedProtocols.UNKNOWN,
                request_size=SIZE_NOT_SET,
                request_time=None,
                response_size=SIZE_NOT_SET,
                response_code=200,
                url='',
                **kw):
        """Invokes the base constructor with default values."""
        op_info = operation.Info(**kw)
        _validate_timedelta_arg('backend_time', backend_time)
        _validate_timedelta_arg('overhead_time', overhead_time)
        _validate_timedelta_arg('request_time', request_time)
        _validate_int_arg('request_size', request_size)
        _validate_int_arg('response_size', response_size)
        if not isinstance(protocol, ReportedProtocols):
            raise ValueError('protocol should be a %s' % (ReportedProtocols,))
        if not isinstance(platform, ReportedPlatforms):
            raise ValueError('platform should be a %s' % (ReportedPlatforms,))
        if not isinstance(error_cause, ErrorCause):
            raise ValueError('error_cause should be a %s' % (ErrorCause,))
        return super(cls, Info).__new__(
            cls,
            api_name,
            api_method,
            api_version,
            auth_issuer,
            auth_audience,
            backend_time,
            error_cause,
            location,
            log_message,
            method,
            overhead_time,
            platform,
            producer_project_id,
            protocol,
            request_size,
            request_time,
            response_code,
            response_size,
            url,
            **op_info._asdict())

    def _as_log_entry(self, name, now):
        """Makes a `LogEntry` from this instance for the given log_name.

        Args:
          rules (:class:`ReportingRules`): determines what labels, metrics and
            logs to include in the report request.
          now (:class:`datetime.DateTime`): the current time

        Return:
          a ``LogEntry`` generated from this instance with the given name
          and timestamp

        Raises:
          ValueError: if the fields in this instance are insufficient to
            to create a valid ``ServicecontrolServicesReportRequest``

        """
        # initialize the struct with fields that are always present
        d = {
            'http_response_code': self.response_code,
            'timestamp': time.mktime(now.timetuple())
        }

        # compute the severity
        severity = _SEVERITY.INFO
        if self.response_code >= 400:
            severity = _SEVERITY.ERROR
            d['error_cause'] = self.error_cause.name

        # add 'optional' fields to the struct
        if self.request_size > 0:
            d['request_size'] = self.request_size
        if self.response_size > 0:
            d['response_size'] = self.response_size
        if self.method:
            d['http_method'] = self.method
        if self.request_time:
            d['request_latency_in_ms'] = self.request_time.total_seconds() * 1000

        # add 'copyable' fields to the struct
        for key in self.COPYABLE_LOG_FIELDS:
            value = getattr(self, key, None)
            if value:
                d[key] = value

        return messages.LogEntry(
            name=name,
            timestamp=timestamp.to_rfc3339(now),
            severity=severity,
            structPayload=_struct_payload_from(d))

    def as_report_request(self, rules, timer=datetime.utcnow):
        """Makes a `ServicecontrolServicesReportRequest` from this instance

        Args:
          rules (:class:`ReportingRules`): determines what labels, metrics and
            logs to include in the report request.
          timer: a function that determines the current time

        Return:
          a ``ServicecontrolServicesReportRequest`` generated from this instance
          governed by the provided ``rules``

        Raises:
          ValueError: if the fields in this instance cannot be used to create
            a valid ``ServicecontrolServicesReportRequest``

        """
        if not self.service_name:
            raise ValueError('the service name must be set')
        op = super(Info, self).as_operation(timer=timer)

        # Populate metrics and labels if they can be associated with a
        # method/operation
        if op.operationId and op.operationName:
            labels = {}
            for known_label in rules.labels:
                known_label.do_labels_update(self, labels)
            if labels:
                op.labels = encoding.PyValueToMessage(
                    messages.Operation.LabelsValue,
                    labels)
            for known_metric in rules.metrics:
                known_metric.do_operation_update(self, op)

        # Populate the log entries
        now = timer()
        op.logEntries = [self._as_log_entry(l, now) for l in rules.logs]

        return messages.ServicecontrolServicesReportRequest(
            serviceName=self.service_name,
            reportRequest=messages.ReportRequest(operations=[op]))


_NO_RESULTS = tuple()


class Aggregator(object):
    """Aggregates Service Control Report requests.

    :func:`report` determines if a `ReportRequest` should be sent to the
    service immediately

    """

    CACHED_OK = object()
    """A sentinel returned by :func:`report` when a request is cached OK."""

    MAX_OPERATION_COUNT = 1000
    """The maximum number of operations to send in a report request."""

    def __init__(self, service_name, options, kinds=None,
                 timer=datetime.utcnow):
        """
        Constructor

        Args:
          service_name (string): name of the service being aggregagated
          options (:class:`google.api.caches.ReportOptions`): configures the behavior
            of this aggregator
          kinds (dict[string, [:class:`.MetricKind`]]): describes the
            type of metrics used during aggregation
          timer (function([[datetime]]): a function that returns the current
            as a time as a datetime instance

        """
        self._cache = caches.create(options, timer=timer)
        self._options = options
        self._kinds = kinds
        self._service_name = service_name

    @property
    def flush_interval(self):
        """The interval between calls to flush.

        Returns:
           timedelta: the period between calls to flush if, or ``None`` if no
           cache is set

        """
        return None if self._cache is None else self._options.flush_interval

    @property
    def service_name(self):
        """The service to which all requests being aggregated should belong."""
        return self._service_name

    def flush(self):
        """Flushes this instance's cache.

        The driver of this instance should call this method every
        `flush_interval`.

        Returns:
          list[``ServicecontrolServicesReportRequest``]: corresponding to the
            pending cached operations

        """
        if self._cache is None:
            return _NO_RESULTS
        with self._cache as c:
            flushed_ops = [x.as_operation() for x in list(c.out_deque)]
            c.out_deque.clear()
            reqs = []
            max_ops = self.MAX_OPERATION_COUNT
            for x in range(0, len(flushed_ops), max_ops):
                report_request = messages.ReportRequest(
                    operations=flushed_ops[x:x + max_ops])
                reqs.append(
                    messages.ServicecontrolServicesReportRequest(
                        serviceName=self.service_name,
                        reportRequest=report_request))

            return reqs

    def clear(self):
        """Clears the cache."""
        if self._cache is None:
            return _NO_RESULTS
        if self._cache is not None:
            with self._cache as k:
                res = [x.as_operation() for x in k.values()]
                k.clear()
                k.out_deque.clear()
                return res

    def report(self, req):
        """Adds a report request to the cache.

        Returns ``None`` if it could not be aggregated, and callers need to
        send the request to the server, otherwise it returns ``CACHED_OK``.

        Args:
           req (:class:`messages.ReportRequest`): the request
             to be aggregated

        Result:
           ``None`` if the request as not cached, otherwise ``CACHED_OK``

        """
        if self._cache is None:
            return None  # no cache, send request now
        if not isinstance(req, messages.ServicecontrolServicesReportRequest):
            raise ValueError('Invalid request')
        if req.serviceName != self.service_name:
            logger.error('bad report(): service_name %s does not match ours %s',
                         req.serviceName, self.service_name)
            raise ValueError('Service name mismatch')
        report_req = req.reportRequest
        if report_req is None:
            logger.error('bad report(): no report_request in %s', req)
            raise ValueError('Expected report_request not set')
        if _has_high_important_operation(report_req) or self._cache is None:
            return None
        ops_by_signature = _key_by_signature(report_req.operations,
                                             _sign_operation)

        # Concurrency:
        #
        # This holds a lock on the cache while updating it.  No i/o operations
        # are performed, so any waiting threads see minimal delays
        with self._cache as cache:
            for key, op in iter(ops_by_signature.items()):
                agg = cache.get(key)
                if agg is None:
                    cache[key] = operation.Aggregator(op, self._kinds)
                else:
                    agg.add(op)

        return self.CACHED_OK


def _has_high_important_operation(req):
    def is_important(op):
        return (op.importance !=
                messages.Operation.ImportanceValueValuesEnum.LOW)

    return functools.reduce(lambda x, y: x and is_important(y),
                            req.operations, True)


def _key_by_signature(operations, signature_func):
    """Creates a dictionary of operations keyed by signature

    Args:
      operations (iterable[Operations]): the input operations

    Returns:
       dict[string, [Operations]]: the operations keyed by signature
    """
    return dict((signature_func(op), op) for op in operations)


def _sign_operation(op):
    """Obtains a signature for an operation in a ReportRequest.

    Args:
       op (:class:`google.api.gen.servicecontrol_v1_messages.Operation`): an
         operation used in a `ReportRequest`

    Returns:
       string: a unique signature for that operation
    """
    md5 = hashlib.md5()
    md5.update(op.consumerId)
    md5.update('\x00')
    md5.update(op.operationName)
    if op.labels:
        signing.add_dict_to_hash(md5, encoding.MessageToPyValue(op.labels))
    return md5.digest()



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import sys
import unittest2
from expects import expect, equal, raise_error

from google.api.control import money, messages


class TestCheckValid(unittest2.TestCase):
    _BAD_CURRENCY = messages.Money(currencyCode='this-is-bad')
    _MISMATCHED_UNITS = (
        messages.Money(currencyCode='JPY', units=-1, nanos=1),
        messages.Money(currencyCode='JPY', units=1, nanos=-1),
    )
    _NANOS_OOB = messages.Money(currencyCode='EUR', units=0, nanos=9999999999)
    _OK = (
        messages.Money(currencyCode='JPY', units=1, nanos=1),
        messages.Money(currencyCode='JPY', units=-1, nanos=-1),
        messages.Money(currencyCode='EUR', units=0, nanos=money.MAX_NANOS),
    )

    def test_should_fail_if_not_really_money(self):
        expect(lambda: money.check_valid(object())).to(raise_error(ValueError))
        expect(lambda: money.check_valid(None)).to(raise_error(ValueError))

    def test_should_fail_when_no_currency_is_set(self):
        expect(lambda: money.check_valid(messages.Money())).to(
            raise_error(ValueError))

    def test_should_fail_when_the_currency_is_bad(self):
        expect(lambda: money.check_valid(self._BAD_CURRENCY)).to(
            raise_error(ValueError))

    def test_should_fail_when_the_units_and_nanos_are_mismatched(self):
        for m in self._MISMATCHED_UNITS:
            expect(lambda: money.check_valid(m)).to(raise_error(ValueError))

    def test_should_fail_when_nanos_are_oob(self):
        expect(lambda: money.check_valid(self._NANOS_OOB)).to(
            raise_error(ValueError))

    def test_should_succeed_for_ok_instances(self):
        for m in self._OK:
            money.check_valid(m)


class TestAdd(unittest2.TestCase):
    _SOME_YEN = messages.Money(currencyCode='JPY', units=3, nanos=0)
    _SOME_YEN_DEBT = messages.Money(currencyCode='JPY', units=-2, nanos=-1)
    _SOME_MORE_YEN = messages.Money(currencyCode='JPY', units=1, nanos=3)
    _SOME_USD = messages.Money(currencyCode='USD', units=1, nanos=0)
    _INT64_MAX = sys.maxint
    _INT64_MIN = -sys.maxint - 1
    _LARGE_YEN = messages.Money(currencyCode='JPY', units=_INT64_MAX -1, nanos=0)
    _LARGE_YEN_DEBT = messages.Money(currencyCode='JPY', units=-_INT64_MAX + 1, nanos=0)

    def test_should_fail_if_non_money_is_used(self):
        testfs = [
            lambda: money.add(self._SOME_YEN, object()),
            lambda: money.add(object(), self._SOME_USD),
            lambda: money.add(None, self._SOME_USD),
            lambda: money.add(self._SOME_YEN, None),
        ]
        for testf in testfs:
            expect(testf).to(raise_error(ValueError))

    def test_should_fail_on_currency_mismatch(self):
        testf = lambda: money.add(self._SOME_YEN, self._SOME_USD)
        expect(testf).to(raise_error(ValueError))

    def test_should_fail_on_unallowed_positive_overflows(self):
        testf = lambda: money.add(self._SOME_YEN, self._LARGE_YEN)
        expect(testf).to(raise_error(OverflowError))

    def test_should_allow_positive_overflows(self):
        overflowing = money.add(self._SOME_YEN, self._LARGE_YEN,
                                allow_overflow=True)
        expect(overflowing.units).to(equal(self._INT64_MAX))
        expect(overflowing.nanos).to(equal(money.MAX_NANOS))

    def test_should_fail_on_unallowed_negative_overflows(self):
        testf = lambda: money.add(self._SOME_YEN_DEBT, self._LARGE_YEN_DEBT)
        expect(testf).to(raise_error(OverflowError))

    def test_should_allow_negative_overflows(self):
        overflowing = money.add(self._SOME_YEN_DEBT, self._LARGE_YEN_DEBT,
                                allow_overflow=True)
        expect(overflowing.units).to(equal(self._INT64_MIN))
        expect(overflowing.nanos).to(equal(-money.MAX_NANOS))

    def test_should_add_ok_when_nanos_have_same_sign(self):
        the_sum = money.add(self._SOME_YEN, self._SOME_YEN)
        expect(the_sum.units).to(equal(2 * self._SOME_YEN.units))

    def test_should_add_ok_when_nanos_have_different_signs(self):
        the_sum = money.add(self._SOME_YEN, self._SOME_YEN_DEBT)
        want_units = self._SOME_YEN_DEBT.units + self._SOME_YEN.units - 1
        expect(the_sum.units).to(equal(want_units))
        expect(the_sum.nanos).to(equal(money.MAX_NANOS))
        the_sum = money.add(self._SOME_MORE_YEN, self._SOME_YEN_DEBT)
        want_units = self._SOME_YEN_DEBT.units + self._SOME_YEN.units - 1
        expect(the_sum.units).to(equal(want_units))
        expect(the_sum.nanos).to(equal(1 - money.MAX_NANOS))
